"""Test fixtures."""
import pytest
import numpy as np
import tensorflow as tf

import aboleth as ab

from scipy.special import expit


SEED = 666
RAND = np.random.RandomState(SEED)


@pytest.fixture
def random():
    return RAND


@pytest.fixture
def make_data():
    """Make some simple data."""
    N = 100
    x1 = np.linspace(-10, 10, N)
    x2 = np.linspace(-1, 1, N)**2
    x = np.vstack((x1, x2)).T
    w = np.array([[0.5], [2.0]])
    Y = np.dot(x, w) + RAND.randn(N, 1)
    X = tf.tile(tf.expand_dims(x, 0), [3, 1, 1])
    return x, Y, X


@pytest.fixture
def make_image_data():
    """Make some simple data."""
    N = 100
    M = 5
    # N 28x28 RGB float images
    x = expit(RAND.randn(N, 28, 28, 3)).astype(np.float32)
    w = np.linspace(-2.5, 2.5, 28*28*3)
    Y = np.dot(x.reshape(-1, 28*28*3), w) + RAND.randn(N, 1)
    X = tf.tile(tf.expand_dims(x, 0), [M, 1, 1, 1, 1])
    return x, Y, X


@pytest.fixture
def make_missing_data():
    """Make some simple data."""
    N = 10
    D = 5
    x = np.ones((N, D)) * np.linspace(1, D, D)
    mask = np.zeros((N, D)).astype(bool)
    mask[N-5:] = True
    x[mask] = 666.
    X = tf.tile(tf.expand_dims(x, 0), [3, 1, 1])
    X = tf.cast(X, tf.float32)
    w = np.linspace(1, 5, 5)[np.newaxis, :]
    y = np.sum(x * w, axis=1)
    return x, mask, X, y


@pytest.fixture
def make_categories():
    """Make some simple categorical data."""
    N = 100
    K = 5
    x = RAND.randint(0, K, size=N)[:, np.newaxis]
    x = x.astype(np.int32)
    return x, K


@pytest.fixture
def make_graph():
    """Make the requirements for making a simple tf graph."""
    x, Y, X = make_data()

    layers = ab.stack(
        ab.InputLayer(name='X', n_samples=10),
        lambda X: (X[:, :, 0:1], 0.0)   # Mock a sampling layer
    )
    N = len(x)

    X_ = tf.placeholder(tf.float32, x.shape)
    Y_ = tf.placeholder(tf.float32, Y.shape)
    N_ = tf.placeholder(tf.float32)

    return x, Y, N, X_, Y_, N_, layers



"""Higher-order neural network layers (made from other layers)."""
import tensorflow as tf

from aboleth.baselayers import Layer, MultiLayer


class Concat(MultiLayer):
    r"""Concatenates the output of multiple layers.

    Parameters
    ----------
    layers : [MultiLayer]
        The layers to concatenate.

    """

    def __init__(self, *layers):
        """Initialise with the individual layers."""
        self.layers = layers

    def _build(self, **kwargs):
        """Build the concatenation."""
        tensors, losses = zip(*map(lambda l: l(**kwargs), self.layers))
        result = tf.concat(tensors, axis=-1)
        loss = tf.add_n(losses)
        return result, loss


class Sum(MultiLayer):
    r"""Sums multiple layers by adding their outputs.

    Parameters
    ----------
    layers : [MultiLayer]
        The layers to add.

    """

    def __init__(self, *layers):
        """Initialise with the individual layers."""
        self.layers = layers

    def _build(self, **kwargs):
        """Build the summation layer."""
        tensors, losses = zip(*map(lambda l: l(**kwargs), self.layers))
        result = tf.add_n(tensors)
        loss = tf.add_n(losses)
        return result, loss


class PerFeature(Layer):
    r"""Concatenate multiple layers with sliced inputs.

    Each layer will recieve a slice along the last axis of the input to the new
    function. In other words, ``PerFeature(l1, l2)(X)``
    will call ``l1(X[..., 0]) and l2(X[..., 1])`` then concatenate their
    outputs into a single tensor. This is mostly useful for simplifying
    embedding multiple categorical inputs that are stored columnwise
    in the same 2D tensor.

    This function assumes the tensor being provided is 3D.

    Parameters
    ----------
    layers : [Layer]
        The layers to concatenate.

    """

    def __init__(self, *layers):
        """Initialise with the individual layers."""
        self.layers = layers

    def _build(self, X):
        """Build slice concatenation operation. ``X`` is a rank 3 Tensor."""
        rank = len(X.shape)
        assert rank == 3

        tensors, losses = zip(*[l(X[..., i:i + 1])
                                for i, l in enumerate(self.layers)])
        result = tf.concat(tensors, axis=-1)
        loss = tf.add_n(losses)
        return result, loss



"""Test the hlayers module."""
import numpy as np
import tensorflow as tf
import aboleth as ab


def test_concat(make_data):
    """Test concatenation layer."""
    x, _, X = make_data

    # This replicates the input layer behaviour
    f = ab.InputLayer('X', n_samples=3)
    g = ab.InputLayer('Y', n_samples=3)

    catlayer = ab.Concat(f, g)

    F, KL = catlayer(X=x, Y=x)

    tc = tf.test.TestCase()
    with tc.test_session():
        forked = F.eval()
        orig = X.eval()
        assert forked.shape == orig.shape[0:2] + (2 * orig.shape[2],)
        assert np.all(forked == np.dstack((orig, orig)))
        assert KL.eval() == 0.0


def test_perfeature(make_data):
    """Test per-feature application of layers."""
    x, _, X = make_data

    def make_idxlayer(i):
        def idlayer(X):
            return X + i, float(i)
        return idlayer

    catlayer = ab.PerFeature(make_idxlayer(2), make_idxlayer(3))
    F, KL = catlayer(X)

    tc = tf.test.TestCase()
    with tc.test_session():
        catted = F.eval()
        orig = X.eval()
        assert catted.shape == orig.shape
        assert np.allclose(catted[:, :, 0], orig[:, :, 0] + 2)
        assert np.allclose(catted[:, :, 1], orig[:, :, 1] + 3)
        assert KL.eval() == 5.0


def test_sum(make_data):
    """Test the summation layer."""
    x, _, X = make_data

    # This replicates the input layer behaviour
    def f(**kwargs):
        return kwargs['X'], 0.0

    def g(**kwargs):
        return kwargs['Y'], 0.0

    addlayer = ab.Sum(f, g)

    F, KL = addlayer(X=X, Y=X)

    tc = tf.test.TestCase()
    with tc.test_session():
        forked = F.eval()
        orig = X.eval()
        assert forked.shape == orig.shape
        assert np.all(forked == 2 * orig)
        assert KL.eval() == 0.0



"""Test the kernels module."""
import pytest
import numpy as np
import tensorflow as tf

import aboleth as ab

kernel_list = [
    (ab.RBF, {}),
    (ab.RBFVariational, {}),
    (ab.Matern, {'p': 1}),
    (ab.Matern, {'p': 2})
]


@pytest.mark.parametrize('kernels', kernel_list)
def test_shift_invariant_kernels(kernels):
    """Test random kernels approximations."""
    d, D = 10, 100
    kern, p = kernels
    k = kern(**p)

    # Check dim
    P, KL = k.weights(input_dim=d, n_features=D)
    assert P.shape == (d, D)


@pytest.mark.parametrize('kernels', kernel_list)
@pytest.mark.parametrize('lenscales', [
    1.0,
    np.array([1.0]),
    np.ones((10, 1), dtype=np.float32),
    tf.constant(1.0),
    tf.ones((10, 1))
])
def test_ARD_lenscales(kernels, lenscales):
    """Test random kernels with multi-dim lenscales."""
    d, D = 10, 100
    kern, p = kernels
    k = kern(lenscale=lenscales, **p)

    # Check dim
    P, KL = k.weights(input_dim=d, n_features=D)
    assert P.shape == (d, D)
